import logging
from os import getenv

from tinydb import TinyDB


class Database:
    """ Class for handling local service information and storing queues """

    def __init__(self, *args):
        """ Accepts any number of strings based on which to create or connect databases """
        self.dbs = {arg: TinyDB(f"{arg}.json") for arg in args}
        self.logger = logging.getLogger("__main__")
        self.logger.setLevel(getenv("LEVEL"))

    def get_id(self, document):
        """ Returns document ID """
        return document.doc_id if hasattr(document, "doc_id") else None

    def remove(self, db, ID):
        """ Removes a document from a db by ID """
        self.dbs[db].remove(doc_ids=[ID])
        self.logger.info(f"Removed document from {db} with id {ID}")

    def insert(self, db, document):
        """ Inserts document and returns it's ID if not already inserted """
        ID = self.get_id(document)
        if ID is None:
            ID = self.dbs[db].insert(document)
            self.logger.info(f"Inserted document to {db} with id {ID}")

        return ID

    def documents(self, db):
        """ Returns all documents from db """
        documents = self.dbs[db].all()
        return documents



import asyncio
from types import FunctionType

from box import Box

from subito.db import queue
from subito.utils import extract
from subito.operations import operation


# TODO: Appropriate decorators inside class - how?
def exceptionHandler(decorated):
    async def wrapper(self, *args, **kwargs):
        try:
            await decorated(self, *args, **kwargs)
        except Exception as e:
            self.logger.error(f"Service not implemented or unavailable for {self.name}")

    return wrapper


def queueHandler(decorated):
    async def wrapper(self, *args, **kwargs):
        if not kwargs:
            return

        ID = self.queue.schedule(kwargs)
        result = await decorated(self, Box(kwargs))
        self.queue.complete(ID)
        return result

    return wrapper


class Subscription(operation.Operation):
    def __init__(self, document: str):
        """ Initiates a subscription monitor. \nIt accepts a document argument which represents a path to a file or a string representing graphql subscription and a handler function to process messages """
        super().__init__(document)
        self._handler = None
        self.queue = queue.Queue(self.name)

    @exceptionHandler
    async def dequeue(self, session):
        """ If messages queued in database process those """
        # async with self.client as session:
        if messages := self.queue.queue():
            await asyncio.gather(*[self.process(**message) for message in messages])

    @exceptionHandler
    async def subscribe(self, session):
        """ Subscribe to a subscription, and process it's messages """
        async for message in session.subscribe(self.ast):
            await self.process(**message)

    @queueHandler
    async def process(self, data):
        """ Send to a service for further processing """
        return await self.handler(data)

    @property
    def handler(self):
        return self._handler

    @handler.setter
    def handler(self, handler):
        self._handler = handler



from asyncio import gather
from types import FunctionType

from gql import Client, WebsocketsTransport

from subito.operations.subscription import Subscription


class Subito:
    session = None
    subscriptions = []

    def __init__(self, url: str):
        self.url = url
        self.options = {
            "transport": WebsocketsTransport(url=self.url),
            "fetch_schema_from_transport": True,
        }

    async def run(self):
        async with Client(**self.options) as session:
            Subito.session = session
            await gather(*(s.dequeue(session) for s in Subito.subscriptions))
            await gather(*(s.subscribe(session) for s in Subito.subscriptions))

    def subscribe(self, document: str):
        Subito.subscriptions.append(Subscription(document))

        def inner(handler):
            subscription = Subito.subscriptions[-1]
            subscription.handler = handler

            async def wrapper(self, *args, **kwargs):
                return handler

            return wrapper

        return inner



""" Store ubiquitous helper functions """
# from time import time
# from os import listdir, getenv, path as p

from box import Box


def extract(message):
    """ Exract the data nad key value from GraphQL query, mutation or subscription result result """
    (key,) = [key for key in message.keys()]
    data = Box(message[key]) if hasattr(message[key], "keys") else message[key]
    return data, key



from subito.db import database


class Queue(database.Database):
    def __init__(self, subscription):
        """ Creates a database object for queing subscription messages """
        super().__init__(subscription)
        self.subscription = subscription

    def queue(self):
        return self.documents(self.subscription)

    def schedule(self, message):
        return self.insert(self.subscription, message)

    def complete(self, message):
        return self.remove(self.subscription, message)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['subito', 'subito.db', 'subito.operations']

package_data = \
{'': ['*']}

install_requires = \
['gql>=3.0.0-alpha.1,<4.0.0',
 'python-box>=5.1.1,<6.0.0',
 'tinydb>=4.1.1,<5.0.0']

setup_kwargs = {
    'name': 'subito',
    'version': '0.1.0',
    'description': 'A libarary which facilitates quickly getting up and running tasks in reaction to one or more GraphQL subscriptions.',
    'long_description': None,
    'author': 'Milos Grujic',
    'author_email': 'srbiotik@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
